"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.completion = exports.commandNames = exports.shorthands = exports.cliOptionsTypes = void 0;
exports.rcOptionsTypes = rcOptionsTypes;
exports.help = help;
exports.handler = handler;
exports.getCellWidth = getCellWidth;
exports.toOutdatedWithVersionDiff = toOutdatedWithVersionDiff;
exports.renderPackageName = renderPackageName;
exports.renderCurrent = renderCurrent;
exports.renderLatest = renderLatest;
exports.renderDetails = renderDetails;
const cli_utils_1 = require("@pnpm/cli-utils");
const colorize_semver_diff_1 = __importDefault(require("@pnpm/colorize-semver-diff"));
const common_cli_options_help_1 = require("@pnpm/common-cli-options-help");
const config_1 = require("@pnpm/config");
const error_1 = require("@pnpm/error");
const outdated_1 = require("@pnpm/outdated");
const semver_diff_1 = __importDefault(require("@pnpm/semver-diff"));
const table_1 = require("@zkochan/table");
const chalk_1 = __importDefault(require("chalk"));
const pick_1 = __importDefault(require("ramda/src/pick"));
const sortWith_1 = __importDefault(require("ramda/src/sortWith"));
const render_help_1 = __importDefault(require("render-help"));
const util_1 = require("util");
const utils_1 = require("./utils");
const recursive_1 = require("./recursive");
function rcOptionsTypes() {
    return {
        ...(0, pick_1.default)([
            'depth',
            'dev',
            'global-dir',
            'global',
            'long',
            'optional',
            'production',
        ], config_1.types),
        compatible: Boolean,
        format: ['table', 'list', 'json'],
        'sort-by': 'name',
    };
}
const cliOptionsTypes = () => ({
    ...rcOptionsTypes(),
    recursive: Boolean,
});
exports.cliOptionsTypes = cliOptionsTypes;
exports.shorthands = {
    D: '--dev',
    P: '--production',
    table: '--format=table',
    'no-table': '--format=list',
    json: '--format=json',
};
exports.commandNames = ['outdated'];
function help() {
    return (0, render_help_1.default)({
        description: `Check for outdated packages. The check can be limited to a subset of the installed packages by providing arguments (patterns are supported).

Examples:
pnpm outdated
pnpm outdated --long
pnpm outdated gulp-* @babel/core`,
        descriptionLists: [
            {
                title: 'Options',
                list: [
                    {
                        description: 'Print only versions that satisfy specs in package.json',
                        name: '--compatible',
                    },
                    {
                        description: 'By default, details about the outdated packages (such as a link to the repo) are not displayed. \
To display the details, pass this option.',
                        name: '--long',
                    },
                    {
                        description: 'Check for outdated dependencies in every package found in subdirectories \
or in every workspace package, when executed inside a workspace. \
For options that may be used with `-r`, see "pnpm help recursive"',
                        name: '--recursive',
                        shortAlias: '-r',
                    },
                    {
                        description: 'Prints the outdated packages in a list. Good for small consoles',
                        name: '--no-table',
                    },
                    {
                        description: 'Check only "dependencies" and "optionalDependencies"',
                        name: '--prod',
                        shortAlias: '-P',
                    },
                    {
                        description: 'Check only "devDependencies"',
                        name: '--dev',
                        shortAlias: '-D',
                    },
                    {
                        description: 'Don\'t check "optionalDependencies"',
                        name: '--no-optional',
                    },
                    {
                        description: 'Prints the outdated dependencies in the given format. Default is "table". Supported options: "table, list, json"',
                        name: '--format <format>',
                    },
                    {
                        description: 'Specify the sorting method. Currently only `name` is supported.',
                        name: '--sort-by',
                    },
                    common_cli_options_help_1.OPTIONS.globalDir,
                    ...common_cli_options_help_1.UNIVERSAL_OPTIONS,
                ],
            },
            common_cli_options_help_1.FILTERING,
        ],
        url: (0, cli_utils_1.docsUrl)('outdated'),
        usages: ['pnpm outdated [<pkg> ...]'],
    });
}
const completion = async (cliOpts) => {
    return (0, cli_utils_1.readDepNameCompletions)(cliOpts.dir);
};
exports.completion = completion;
async function handler(opts, params = []) {
    const include = {
        dependencies: opts.production !== false,
        devDependencies: opts.dev !== false,
        optionalDependencies: opts.optional !== false,
    };
    if (opts.recursive && (opts.selectedProjectsGraph != null)) {
        const pkgs = Object.values(opts.selectedProjectsGraph).map((wsPkg) => wsPkg.package);
        return (0, recursive_1.outdatedRecursive)(pkgs, params, { ...opts, include });
    }
    const manifest = await (0, cli_utils_1.readProjectManifestOnly)(opts.dir, opts);
    const packages = [
        {
            rootDir: opts.dir,
            manifest,
        },
    ];
    const [outdatedPackages] = await (0, outdated_1.outdatedDepsOfProjects)(packages, params, {
        ...opts,
        fullMetadata: opts.long,
        ignoreDependencies: opts.updateConfig?.ignoreDependencies,
        include,
        retry: {
            factor: opts.fetchRetryFactor,
            maxTimeout: opts.fetchRetryMaxtimeout,
            minTimeout: opts.fetchRetryMintimeout,
            retries: opts.fetchRetries,
        },
        timeout: opts.fetchTimeout,
    });
    let output;
    switch (opts.format ?? 'table') {
        case 'table': {
            output = renderOutdatedTable(outdatedPackages, opts);
            break;
        }
        case 'list': {
            output = renderOutdatedList(outdatedPackages, opts);
            break;
        }
        case 'json': {
            output = renderOutdatedJSON(outdatedPackages, opts);
            break;
        }
        default: {
            throw new error_1.PnpmError('BAD_OUTDATED_FORMAT', `Unsupported format: ${opts.format?.toString() ?? 'undefined'}`);
        }
    }
    return {
        output,
        exitCode: outdatedPackages.length === 0 ? 0 : 1,
    };
}
function renderOutdatedTable(outdatedPackages, opts) {
    if (outdatedPackages.length === 0)
        return '';
    const columnNames = [
        'Package',
        'Current',
        'Latest',
    ];
    const columnFns = [
        renderPackageName,
        renderCurrent,
        renderLatest,
    ];
    if (opts.long) {
        columnNames.push('Details');
        columnFns.push(renderDetails);
    }
    // Avoid the overhead of allocating a new array caused by calling `array.map()`
    for (let i = 0; i < columnNames.length; i++)
        columnNames[i] = chalk_1.default.blueBright(columnNames[i]);
    const data = [
        columnNames,
        ...sortOutdatedPackages(outdatedPackages, { sortBy: opts.sortBy })
            .map((outdatedPkg) => columnFns.map((fn) => fn(outdatedPkg))),
    ];
    let detailsColumnMaxWidth = 40;
    if (opts.long) {
        detailsColumnMaxWidth = outdatedPackages.filter(pkg => pkg.latestManifest && !pkg.latestManifest.deprecated).reduce((maxWidth, pkg) => {
            const cellWidth = pkg.latestManifest?.homepage?.length ?? 0;
            return Math.max(maxWidth, cellWidth);
        }, 0);
    }
    return (0, table_1.table)(data, {
        ...cli_utils_1.TABLE_OPTIONS,
        columns: {
            ...cli_utils_1.TABLE_OPTIONS.columns,
            // Detail column:
            3: {
                width: detailsColumnMaxWidth,
                wrapWord: true,
            },
        },
    });
}
function renderOutdatedList(outdatedPackages, opts) {
    if (outdatedPackages.length === 0)
        return '';
    return sortOutdatedPackages(outdatedPackages, { sortBy: opts.sortBy })
        .map((outdatedPkg) => {
        let info = `${chalk_1.default.bold(renderPackageName(outdatedPkg))}
${renderCurrent(outdatedPkg)} ${chalk_1.default.grey('=>')} ${renderLatest(outdatedPkg)}`;
        if (opts.long) {
            const details = renderDetails(outdatedPkg);
            if (details) {
                info += `\n${details}`;
            }
        }
        return info;
    })
        .join('\n\n') + '\n';
}
function renderOutdatedJSON(outdatedPackages, opts) {
    const outdatedPackagesJSON = sortOutdatedPackages(outdatedPackages, { sortBy: opts.sortBy })
        .reduce((acc, outdatedPkg) => {
        acc[outdatedPkg.packageName] = {
            current: outdatedPkg.current,
            latest: outdatedPkg.latestManifest?.version,
            wanted: outdatedPkg.wanted,
            isDeprecated: Boolean(outdatedPkg.latestManifest?.deprecated),
            dependencyType: outdatedPkg.belongsTo,
        };
        if (opts.long) {
            acc[outdatedPkg.packageName].latestManifest = outdatedPkg.latestManifest;
        }
        return acc;
    }, {});
    return JSON.stringify(outdatedPackagesJSON, null, 2);
}
function sortOutdatedPackages(outdatedPackages, opts) {
    const sortBy = opts?.sortBy;
    const comparators = (sortBy === 'name') ? [utils_1.NAME_COMPARATOR] : utils_1.DEFAULT_COMPARATORS;
    return (0, sortWith_1.default)(comparators, outdatedPackages.map(toOutdatedWithVersionDiff));
}
function getCellWidth(data, columnNumber, maxWidth) {
    const maxCellWidth = data.reduce((cellWidth, row) => {
        const cellLines = (0, util_1.stripVTControlCharacters)(row[columnNumber]).split('\n');
        const currentCellWidth = cellLines.reduce((lineWidth, line) => {
            return Math.max(lineWidth, line.length);
        }, 0);
        return Math.max(cellWidth, currentCellWidth);
    }, 0);
    return Math.min(maxWidth, maxCellWidth);
}
function toOutdatedWithVersionDiff(outdated) {
    if (outdated.latestManifest != null) {
        return {
            ...outdated,
            ...(0, semver_diff_1.default)(outdated.wanted, outdated.latestManifest.version),
        };
    }
    return {
        ...outdated,
        change: 'unknown',
    };
}
function renderPackageName({ belongsTo, packageName }) {
    switch (belongsTo) {
        case 'devDependencies': return `${packageName} ${chalk_1.default.dim('(dev)')}`;
        case 'optionalDependencies': return `${packageName} ${chalk_1.default.dim('(optional)')}`;
        default: return packageName;
    }
}
function renderCurrent({ current, wanted }) {
    const output = current ?? 'missing';
    if (current === wanted)
        return output;
    return `${output} (wanted ${wanted})`;
}
function renderLatest(outdatedPkg) {
    const { latestManifest, change, diff } = outdatedPkg;
    if (latestManifest == null)
        return '';
    if (change === null || (diff == null)) {
        return latestManifest.deprecated
            ? chalk_1.default.redBright.bold('Deprecated')
            : latestManifest.version;
    }
    return (0, colorize_semver_diff_1.default)({ change, diff });
}
function renderDetails({ latestManifest }) {
    if (latestManifest == null)
        return '';
    const outputs = [];
    if (latestManifest.deprecated) {
        outputs.push(chalk_1.default.redBright(latestManifest.deprecated));
    }
    if (latestManifest.homepage) {
        outputs.push(chalk_1.default.underline(latestManifest.homepage));
    }
    return outputs.join('\n');
}
//# sourceMappingURL=outdated.js.map